﻿Imports System.Reflection
Imports System.Data.Common
Imports System.ComponentModel
Imports System.Windows.Forms
Imports System.Text.RegularExpressions
Imports System.Collections.ObjectModel

Namespace System.Data

   ''' <summary>
   ''' manager-class for a dataset. associates TableAdapters with Datatables. 
   ''' TableAdapters know, how to store and fill a DataTable
   ''' </summary>
   Public Class DatasetAdapter : Inherits DatasetAdapterBase
      Public ReadOnly Provider As String
      Public ReadOnly Connection As DbConnection

      Friend ReadOnly Factory As DbProviderFactory
      Friend ReadOnly Adapters As New Dictionary(Of DataTable, TableAdapter)
      Friend ReadOnly CommandBuilder As DbCommandBuilder
      Friend ReadOnly Quote As Func(Of String, String)
      Friend ReadOnly UnQuoteAll As Func(Of String, String)

      Private ReadOnly _FillJobs As New List(Of TableAdapter.FillJob)
      Private ReadOnly _ParamContainer As DbCommand
      Private ReadOnly _rgxGetSqlWords As Regex
      Private ReadOnly _IdentityCommand As DbCommand
      Private ReadOnly _AddWithValue As Func(Of DbParameterCollection, String, Object, DbParameter)

      Private Shared ReadOnly _rgxJoining As New Regex("^\W*(\w+\W*){2}join", _
         RegexOptions.IgnoreCase)
      Private Shared ReadOnly _SqlServer_MySql As String() = New String() { _
         "System.Data.SqlClient", "MySql.Data.MySqlClient"}

      Public Sub New(ByVal factory As DbProviderFactory, ByVal sConnection As String, ByVal conflictOption As ConflictOption)
         Me.Factory = factory
         Connection = factory.CreateConnection
         Connection.ConnectionString = sConnection
         Provider = Connection.GetType.Namespace
         _ParamContainer = factory.CreateCommand
         Dim tpParams = _ParamContainer.Parameters.GetType
         Dim addWithValue = tpParams.GetMethod("AddWithValue")
         Me._AddWithValue = Function(params, name, value) _
            DirectCast(addWithValue.Invoke(params, New Object() {name, value}), DbParameter)
         CommandBuilder = factory.CreateCommandBuilder
         With CommandBuilder
            .ConflictOption = conflictOption
            .SetAllValues = True
            If .QuotePrefix.Length = 0 Then
               .QuotePrefix = "["
               .QuoteSuffix = "]"
            End If
            Dim sPat = "(?<Rename>\s+As\s+)|(?<Operator>[<=>]{1,2}|\bLIKE\b|\bIN\b)|\?|[\w\.\" _
               .And(.QuotePrefix, "\", .QuoteSuffix, "]+") 'pattern für gequotete Namen anhängen
            _rgxGetSqlWords = New Regex(sPat, RegexOptions.IgnoreCase)
            Dim unquotes = (.QuotePrefix & .QuoteSuffix).ToCharArray
            Quote = AddressOf .QuoteIdentifier
            UnQuoteAll = Function(s) String.Concat(s.Split(unquotes))
         End With
         If Not _SqlServer_MySql.Contains(Provider) Then
            'Access, SqlCe schießen per Event hinter jedem Insert einen Select hinterher
            _IdentityCommand = Connection.CreateCommand
            _IdentityCommand.CommandText = "SELECT @@IDENTITY"
         End If
      End Sub

      Public Overrides Property DataSet() As DataSet
         Get
            Return _DataSet
         End Get
         Set(ByVal value As DataSet)
            If value Is _DataSet Then Return
            AddHandler value.Disposed, AddressOf _Datset_Disposed
            MyBase.DataSet = value
            Adapters.Clear()
            Using Connection.EnterOpen
               Dim cnt = New Counter
               For Each tb As DataTable In _DataSet.GetRankedTables
                  Dim ta = New TableAdapter(tb, Me, cnt.Down)
                  AddHandler ta.RowUpdated, AddressOf _Adapter_RowUpdated
                  Adapters(tb) = ta
               Next
            End Using
         End Set
      End Property

      Friend Function ColumnNames(ByVal table As DataTable) As IEnumerable(Of String)
         Return From p In _ParamContainer.Parameters.Cast(Of DbParameter)() _
            Where p.ParameterName.StartsWith(table.TableName) Select p.SourceColumn
      End Function

      Friend Sub AddParam(ByVal table As DataTable, ByVal param As DbParameter)
         param = param.Clone
         Dim s = String.Concat(table.TableName, ".", param.SourceColumn)
         param.ParameterName = s
         If Not _ParamContainer.Parameters.Contains(s) Then _ParamContainer.Parameters.Add(param)
      End Sub

      Friend Sub AddFillJobs(ByVal jobs As IEnumerable(Of TableAdapter.FillJob))
         '_FillJobs is a kinda consumer-Heap: Inserting first job will start executing. Execution may cause adding of more jobs (to fill childTables), wich will be executed too - respecting their Rank.
         With _FillJobs
            Dim start = .Count = 0
            For Each job In jobs
               Dim i = .BinarySearch(job) Xor -1
               If i < 0 Then Throw Me.Exception("don't add 2 Jobs filling the same Table")
               .Insert(i, job)
            Next
            If Not start Then Return
            Using DataSet.Connection.EnterOpen
               While .Count > 0
                  .Pop.Execute()
               End While
            End Using
         End With
      End Sub

      Friend Function CreateSelectCommand(ByVal Table As String, ByVal distinct As Boolean, ByVal conditions As String, ByVal paramValues As IEnumerable) As DbCommand
         Dim oValues = paramValues.GetEnumerator
         Dim selector = If(distinct, "SELECT DISTINCT ", "SELECT ") _
            .And(Quote(Table), ".* FROM ")
         If Not _rgxJoining.IsMatch(conditions) Then
            'conditions ist keine Join-Anweisung -> TabellenNamen an den selector hängen
            selector &= Quote(Table) & " "
         End If
         Dim cmd = Connection.CreateCommand
         'dummi-matches zufügen, denn die Suche greift 2 zurück oder 1 vor
         Dim mts = _rgxGetSqlWords.Matches("d d ".And(conditions, " d"))
         Dim renames As New Dictionary(Of String, String)
         Dim getColkey = Function(i As Integer) UnQuoteAll(mts(i).Value)
         Dim cnt = New Counter
         For i = 2 To mts.Count - 1
            'Schleife mit 2 Analysen: a) Umbenennungen merken, b) DBParameter erstellen aus Parameter-Platzhalter und paramValues
            If mts(i).Groups("Rename").Success Then ' Umbenennung merken
               renames.Add(getColkey(i + 1), getColkey(i - 1))
            ElseIf mts(i).Value = "?" Then ' DBParameter erstellen
               Dim before = mts(i - 1)
               Dim colKey = ""
               If before.Groups("Operator").Success Then 'column ist links vom '?'
                  colKey = getColkey(i - 2)
               ElseIf before.Value = "?" Then 'IN-Operator: mehrere '?' hintereinander
                  'als colKey den grad vorher zugefügten nehmen
                  colKey = cmd.Parameters(cmd.Parameters.Count - 1).SourceColumn
               ElseIf mts(i + 1).Groups("Operator").Success Then 'column rechts vom '?'
                  colKey = getColkey(i + 2)
               End If
               If Not colKey.Contains(".") Then colKey = Table.And(".", colKey)
               If _ParamContainer.Parameters.Contains(colKey) Then
                  Dim param = _ParamContainer.Parameters(colKey).Clone
                  param.ParameterName = "@p" & cnt.Up
                  param.Value = oValues.Yield
                  cmd.Parameters.Add(param)
               Else 'unknown Sql-Function: hopefully AddWithValue will generate valid Param
                  _AddWithValue(cmd.Parameters, "@p" & cnt.Up, oValues.Yield)
               End If
            End If
         Next
         If oValues.MoveNext Then Throw Me.Exception( _
            "Es wurden mehr Werte übergeben, als Platzhalter im Sql-String angegeben sind")
         If _SqlServer_MySql.Contains(Provider) Then
            'SqlServer verlangt die Parameter-namen - '?' geht nicht. 
            Dim splits = conditions.Split("?"c)
            For i = 0 To cmd.Parameters.Count - 1
               splits(i) &= cmd.Parameters(i).ParameterName
            Next
            conditions = String.Concat(splits)
         End If
         cmd.CommandText = selector & conditions
         Return cmd
      End Function

      Protected Overrides Sub _Save(ByVal tbls As List(Of DataTable))
         'Löschweitergabe zur Erhaltung der Datenkonsistenz: Löschen einer übergeordneten
         ' Datarow löscht alle ChildRows mit
         'Problem beim Update: Update ich den DELETE einer übergeordneten Datarow, so wendet
         ' auch die Datenbank Löschweitergabe auf ihre Daten an. Update ich danach die DELETEs
         ' der ChildRows, so wirft die DB eine Exception, denn bei ihr sind diese Daten schon weg.
         '(Hier vorgetragene) Lösung: Zunächst nur die DELETEs aller übergeordneter Rows des
         ' ganzen Datasets updaten (Top -> Down). Dabei temporär für alle Datarelations des
         ' Datasets AcceptRejectRule.Cascade einstellen. Wenn nun die übergeordneten geupdatet
         ' sind, und (intern) ihr AcceptChanges() aufrufen, akzeptieren die ChildRows auch.
         'Eine akzeptierte (Child-)Datarow gibt mit DataRowState.Unchanged an, daß sie in 
         'Übereinstimmung mit der Datenbank ist, und vom Update auszunehmen.
         Using DataSet.Connection.EnterOpen
            'alle AcceptRejectRules auf Cascade setzen
            Dim switcher = New _AcceptruleSwitcher(DataSet)
            For Each tb In tbls
               If tb.ChildRelations.Count > 0 Then
                  Dim deletes = tb.Select("", "", DataViewRowState.Deleted)
                  If deletes.Count > 0 Then tb.Adapter.Update(deletes)
               End If
            Next
            switcher.Restore()
            'INSERTs und UPDATEs normal updaten
            For Each tb In tbls
               tb.Adapter.Update()
            Next
         End Using
      End Sub
      Private Class _AcceptruleSwitcher
         Inherits List(Of [Set](Of ForeignKeyConstraint, AcceptRejectRule))
         Public Sub New(ByVal dts As DataSet)
            For Each rl As DataRelation In dts.Relations
               Dim ck = rl.ChildKeyConstraint
               If ck.Null Then Continue For
               MyBase.Add(NewSet(ck, ck.AcceptRejectRule))
               ck.AcceptRejectRule = AcceptRejectRule.Cascade
            Next
         End Sub
         Public Sub Restore()
            For Each aSet In Me
               aSet.A.AcceptRejectRule = aSet.B
            Next
         End Sub
      End Class

      Private Sub _Adapter_RowUpdated(ByVal sender As Object, ByVal e As RowUpdatedEventArgs)
         If e.StatementType <> StatementType.Insert Then Return
         _IdentityCommand.Transaction = DirectCast(e.Command.Transaction, DbTransaction)
         Dim primCol = e.Row.Table.PrimaryKey(0)
         Dim id = Convert.ChangeType(_IdentityCommand.ExecuteScalar, primCol.DataType)
         e.Row(primCol) = id
      End Sub

      Private Sub _Datset_Disposed(ByVal s As Object, ByVal e As EventArgs)
         For Each d In New IDisposable() { _
               Connection, CommandBuilder, _ParamContainer, _IdentityCommand}
            If d.NotNull Then d.Dispose()
         Next
         For Each d In Adapters.Keys : d.Dispose() : Next
      End Sub
   End Class

End Namespace