﻿Imports System.Collections.Generic
Imports System.Linq
Imports System.Data.Common
Imports System.ComponentModel
Imports System.Reflection
Imports System.Windows.Forms
Imports System.Text
Imports TableInf = System.[Set](Of System.Data.DataTable, System.IDisposable)

Namespace System.Data

   Partial Public Class DatasetAdapter

      Partial Public Class TableAdapter

         Public WithEvents Adapter As DbDataAdapter
         Public ReadOnly Table As DataTable
         Public ReadOnly AllCommands As New List(Of DbCommand)()
         Private _ChildTables As DataTable() 'in TopDown-Order
         Private _EventEnabler As ActionFrame

#Region "FromCrackedAdapter"

         Public Shared Function FromCrackedAdapter( _
               ByVal crackedAdapter As Component, ByVal dataset As DataSet) As TableAdapter
            Return New TableAdapter(crackedAdapter, dataset)
         End Function

         Private Sub New(ByVal crackedAdapter As Component, ByVal dataset As DataSet)
            Const PropFlag As BindingFlags _
               = BindingFlags.NonPublic Or BindingFlags.Instance Or BindingFlags.GetProperty
            Dim tpToCrack As Type = crackedAdapter.[GetType]()
            Try
               Dim inf As PropertyInfo = tpToCrack.GetProperty("Adapter", PropFlag)
               Adapter = DirectCast(inf.GetValue(crackedAdapter, Nothing), DbDataAdapter)
               Table = dataset.Tables(Adapter.TableMappings(0).DataSetTable)
               inf = tpToCrack.GetProperty("CommandCollection", PropFlag)
               Dim Commands = DirectCast(inf.GetValue(crackedAdapter, Nothing), DbCommand())
               AllCommands.AddRange(Commands)
            Catch x As Exception
#If DEBUG Then
               Stop
#End If
               Throw Me.Exception(x, "kann aus '", crackedAdapter.[GetType](), _
                                  "' nicht die erforderlichen Informationen extrahieren.")
            End Try
            EndInit()
         End Sub

#End Region 'FromCrackedAdapter

#Region "FromTable"

         'ist wesentlich frickeliger als per Reflection, weils unterschiedliche Provider gibt

         Public Shared Function FromTable( _
               ByVal Table As DataTable, ByVal con As DbConnection, _
               ByVal conflictOption As ConflictOption) As TableAdapter
            Return New TableAdapter(Table, con, conflictOption)
         End Function

         Private Sub New( _
               ByVal tb As DataTable, ByVal con As DbConnection, ByVal conflictOption As ConflictOption)
            Table = tb
            Dim columns = (From col In Table.Columns.Cast(Of DataColumn)() _
                           Where col.Expression.HasNoValue _
                           Select col.ColumnName).ToArray
            Dim factory = GetFactory(con)
            'der Commandbuilder öffnet die Connection des Select-Commands und liest die Tabellendefinition aus. Damit konstruiert er delete/insert/update - Commands. Solange er nicht disposed liefert er dem Adapter alle Commands für jeden Datensatz einzeln. Beim Disposen setzt er im Adapter "seine" delete/insert/updates wieder auf Nothing - deshalb verwendich hier Clone.
            Using builder = factory.CreateCommandBuilder
               builder.ConflictOption = conflictOption
               builder.SetAllValues = True
               Dim cmd = con.CreateCommand
               cmd.CommandText.set("SELECT ", ", ".Between(columns), " FROM ", tb.TableName)
               AllCommands.Add(cmd.Clone)
               Adapter = factory.CreateDataAdapter
               With Adapter
                  .SelectCommand = cmd
                  builder.DataAdapter = Adapter
                  .DeleteCommand = builder.GetDeleteCommand.Clone
                  .UpdateCommand = builder.GetUpdateCommand.Clone
                  .InsertCommand = builder.GetInsertCommand.Clone
                  If .InsertCommand.Parameters.Count < columns.Length Then InsertPrim()
                  For i = 0 To columns.Length - 1
                     .InsertCommand.Parameters(i).ParameterName = columns(i)
                  Next
                  .TableMappings.Add(DbDataAdapter.DefaultSourceTableName, Table.TableName)
               End With
            End Using
            EndInit()
         End Sub

         Private Shared Function GetFactory(ByVal con As DbConnection) As DbProviderFactory
            If TypeOf con Is OleDb.OleDbConnection Then
               Return OleDb.OleDbFactory.Instance
            ElseIf TypeOf con Is SqlClient.SqlConnection Then
               Return SqlClient.SqlClientFactory.Instance
            Else
#If DEBUG Then
               Stop
#End If
               Throw New ArgumentException("unknown DBProvider")
            End If
         End Function

         Private Sub InsertPrim()
            With Adapter
               'wenn der commandBuilder in der DB eine serverseitig incrementierende Primkey-Column entdeckt, nimmter die vom Insert aus. Wir frickeln die aber wieder hinein, weil unser AutoIncrement clientseitig ist.
               Dim sPrim = Table.PrimaryKey(0).ColumnName
               'im CommandText öffnet die 1. "(" auf die Spaltennamen, und die 2. "(" auf die Parameter
               Dim splitteds = .InsertCommand.CommandText.Split(New Char() {"("c}, 3)
               'SQLServer hat Spaltennamen in "[]", und Parameter benamt, wo OleDb nur ein "?" hat.
               If TypeOf Adapter Is SqlClient.SqlDataAdapter Then
                  splitteds(0) &= "([" & sPrim & "]"
                  splitteds(1) &= "(@" & sPrim
               ElseIf TypeOf Adapter Is OleDb.OleDbDataAdapter Then
                  splitteds(0) &= "(" & sPrim
                  splitteds(1) &= "(?"
               Else
#If DEBUG Then
                  Stop
#End If
                  Throw New ArgumentException("unknown DBProvider")
               End If
               .InsertCommand.CommandText = ", ".Between(splitteds)
               .InsertCommand.Parameters.Insert(0, .DeleteCommand.Parameters(0).Clone)
            End With
         End Sub

#End Region 'FromTable

         Private Sub EndInit()
            With Adapter
               AllCommands.AddRange(NewArray(.DeleteCommand, .InsertCommand, .UpdateCommand) _
                  .Where(Function(cmd) cmd IsNot Nothing))
            End With
            _ChildTables = (New DataTable() {Table}) _
               .Concat(Table.GetChildRelations.Select(Function(rl) rl.ChildTable)).ToArray
         End Sub

      End Class 'Partial Class TableAdapter

   End Class 'Partial Class DatasetAdapter

End Namespace
